@@ -13,6 +13,7 @@
 #include <linux/if.h>
 #include <net/genetlink.h>
 #include <net/sock.h>
+#include <crypto/algapi.h>
 
 static struct genl_family genl_family;
 
@@ -90,12 +91,21 @@ static int get_allowedips(struct sk_buff *skb, const u8 *ip, u8 cidr,
 	return 0;
 }
 
+struct dump_ctx {
+	struct wg_device *wg;
+	struct wg_peer *next_peer;
+	struct allowedips_node *next_allowedip;
+	u64 allowedips_seq;
+};
+
+#define DUMP_CTX(cb) ((struct dump_ctx *)(cb)->args)
+
 static int
-get_peer(struct wg_peer *peer, struct allowedips_node **next_allowedips_node,
-	 u64 *allowedips_seq, struct sk_buff *skb)
+get_peer(struct wg_peer *peer, struct sk_buff *skb, struct dump_ctx *ctx)
 {
+
 	struct nlattr *allowedips_nest, *peer_nest = nla_nest_start(skb, 0);
-	struct allowedips_node *allowedips_node = *next_allowedips_node;
+	struct allowedips_node *allowedips_node = ctx->next_allowedip;
 	bool fail;
 
 	if (!peer_nest)
@@ -151,9 +161,9 @@ get_peer(struct wg_peer *peer, struct allowedips_node **next_allowedips_node,
 	}
 	if (!allowedips_node)
 		goto no_allowedips;
-	if (!*allowedips_seq)
-		*allowedips_seq = peer->device->peer_allowedips.seq;
-	else if (*allowedips_seq != peer->device->peer_allowedips.seq)
+	if (!ctx->allowedips_seq)
+		ctx->allowedips_seq = peer->device->peer_allowedips.seq;
+	else if (ctx->allowedips_seq != peer->device->peer_allowedips.seq)
 		goto no_allowedips;
 
 	allowedips_nest = nla_nest_start(skb, WGPEER_A_ALLOWEDIPS);
@@ -169,15 +179,15 @@ get_peer(struct wg_peer *peer, struct allowedips_node **next_allowedips_node,
 		if (get_allowedips(skb, ip, cidr, family)) {
 			nla_nest_end(skb, allowedips_nest);
 			nla_nest_end(skb, peer_nest);
-			*next_allowedips_node = allowedips_node;
+			ctx->next_allowedip = allowedips_node;
 			return -EMSGSIZE;
 		}
 	}
 	nla_nest_end(skb, allowedips_nest);
 no_allowedips:
 	nla_nest_end(skb, peer_nest);
-	*next_allowedips_node = NULL;
-	*allowedips_seq = 0;
+	ctx->next_allowedip = NULL;
+	ctx->allowedips_seq = 0;
 	return 0;
 err:
 	nla_nest_cancel(skb, peer_nest);
@@ -197,26 +207,24 @@ static int wg_get_device_start(struct netlink_callback *cb)
 	wg = lookup_interface(attrs, cb->skb);
 	if (IS_ERR(wg))
 		return PTR_ERR(wg);
-	cb->args[0] = (long)wg;
+	DUMP_CTX(cb)->wg = wg;
 	return 0;
 }
 
 static int wg_get_device_dump(struct sk_buff *skb, struct netlink_callback *cb)
 {
-	struct wg_peer *peer, *next_peer_cursor, *last_peer_cursor;
+	struct wg_peer *peer, *next_peer_cursor;
+	struct dump_ctx *ctx = DUMP_CTX(cb);
+	struct wg_device *wg = ctx->wg;
 	struct nlattr *peers_nest;
-	struct wg_device *wg;
 	int ret = -EMSGSIZE;
 	bool done = true;
 	void *hdr;
 
-	wg = (struct wg_device *)cb->args[0];
-	next_peer_cursor = (struct wg_peer *)cb->args[1];
-	last_peer_cursor = (struct wg_peer *)cb->args[1];
-
 	rtnl_lock();
 	mutex_lock(&wg->device_update_lock);
 	cb->seq = wg->device_update_gen;
+	next_peer_cursor = ctx->next_peer;
 
 	hdr = genlmsg_put(skb, NETLINK_CB(cb->skb).portid, cb->nlh->nlmsg_seq,
 			  &genl_family, NLM_F_MULTI, WG_CMD_GET_DEVICE);
@@ -224,7 +232,7 @@ static int wg_get_device_dump(struct sk_buff *skb, struct netlink_callback *cb)
 		goto out;
 	genl_dump_check_consistent(cb, hdr);
 
-	if (!last_peer_cursor) {
+	if (!ctx->next_peer) {
 		if (nla_put_u16(skb, WGDEVICE_A_LISTEN_PORT,
 				wg->incoming_port) ||
 		    nla_put_u32(skb, WGDEVICE_A_FWMARK, wg->fwmark) ||
@@ -257,15 +265,14 @@ static int wg_get_device_dump(struct sk_buff *skb, struct netlink_callback *cb)
 	 * coherent dump anyway, so they'll try again.
 	 */
 	if (list_empty(&wg->peer_list) ||
-	    (last_peer_cursor && list_empty(&last_peer_cursor->peer_list))) {
+	    (ctx->next_peer && list_empty(&ctx->next_peer->peer_list))) {
 		nla_nest_cancel(skb, peers_nest);
 		goto out;
 	}
 	lockdep_assert_held(&wg->device_update_lock);
-	peer = list_prepare_entry(last_peer_cursor, &wg->peer_list, peer_list);
+	peer = list_prepare_entry(ctx->next_peer, &wg->peer_list, peer_list);
 	list_for_each_entry_continue(peer, &wg->peer_list, peer_list) {
-		if (get_peer(peer, (struct allowedips_node **)&cb->args[2],
-			     (u64 *)&cb->args[4] /* and args[5] */, skb)) {
+		if (get_peer(peer, skb, ctx)) {
 			done = false;
 			break;
 		}
@@ -276,7 +283,7 @@ static int wg_get_device_dump(struct sk_buff *skb, struct netlink_callback *cb)
 out:
 	if (!ret && !done && next_peer_cursor)
 		wg_peer_get(next_peer_cursor);
-	wg_peer_put(last_peer_cursor);
+	wg_peer_put(ctx->next_peer);
 	mutex_unlock(&wg->device_update_lock);
 	rtnl_unlock();
 
@@ -286,10 +293,10 @@ out:
 	}
 	genlmsg_end(skb, hdr);
 	if (done) {
-		cb->args[1] = 0;
+		ctx->next_peer = NULL;
 		return 0;
 	}
-	cb->args[1] = (long)next_peer_cursor;
+	ctx->next_peer = next_peer_cursor;
 	return skb->len;
 
 	/* At this point, we can't really deal ourselves with safely zeroing out
@@ -300,12 +307,11 @@ out:
 
 static int wg_get_device_done(struct netlink_callback *cb)
 {
-	struct wg_device *wg = (struct wg_device *)cb->args[0];
-	struct wg_peer *peer = (struct wg_peer *)cb->args[1];
+	struct dump_ctx *ctx = DUMP_CTX(cb);
 
-	if (wg)
-		dev_put(wg->dev);
-	wg_peer_put(peer);
+	if (ctx->wg)
+		dev_put(ctx->wg->dev);
+	wg_peer_put(ctx->next_peer);
 	return 0;
 }
 
@@ -368,8 +374,12 @@ static int set_peer(struct wg_device *wg, struct nlattr **attrs)
 	if (attrs[WGPEER_A_PRESHARED_KEY] &&
 	    nla_len(attrs[WGPEER_A_PRESHARED_KEY]) == NOISE_SYMMETRIC_KEY_LEN)
 		preshared_key = nla_data(attrs[WGPEER_A_PRESHARED_KEY]);
+
 	if (attrs[WGPEER_A_FLAGS])
 		flags = nla_get_u32(attrs[WGPEER_A_FLAGS]);
+	ret = -EOPNOTSUPP;
+	if (flags & ~__WGPEER_F_ALL)
+		goto out;
 
 	ret = -EPFNOSUPPORT;
 	if (attrs[WGPEER_A_PROTOCOL_VERSION]) {
@@ -403,10 +413,16 @@ static int set_peer(struct wg_device *wg, struct nlattr **attrs)
 		}
 		up_read(&wg->static_identity.lock);
 
-		ret = -ENOMEM;
 		peer = wg_peer_create(wg, public_key, preshared_key);
-		if (!peer)
+		if (IS_ERR(peer)) {
+			/* Similar to the above, if the key is invalid, we skip
+			 * it without fanfare, so that services don't need to
+			 * worry about doing key validation themselves.
+			 */
+			ret = PTR_ERR(peer) == -EKEYREJECTED ? 0 : PTR_ERR(peer);
+			peer = NULL;
 			goto out;
+		}
 		/* Take additional reference, as though we've just been
 		 * looked up.
 		 */
@@ -487,6 +503,7 @@ out:
 static int wg_set_device(struct sk_buff *skb, struct genl_info *info)
 {
 	struct wg_device *wg = lookup_interface(info->attrs, skb);
+	u32 flags = 0;
 	int ret;
 
 	if (IS_ERR(wg)) {
@@ -497,6 +514,12 @@ static int wg_set_device(struct sk_buff *skb, struct genl_info *info)
 	rtnl_lock();
 	mutex_lock(&wg->device_update_lock);
 
+	if (info->attrs[WGDEVICE_A_FLAGS])
+		flags = nla_get_u32(info->attrs[WGDEVICE_A_FLAGS]);
+	ret = -EOPNOTSUPP;
+	if (flags & ~__WGDEVICE_F_ALL)
+		goto out;
+
 	ret = -EPERM;
 	if ((info->attrs[WGDEVICE_A_LISTEN_PORT] ||
 	     info->attrs[WGDEVICE_A_FWMARK]) &&
@@ -520,9 +543,7 @@ static int wg_set_device(struct sk_buff *skb, struct genl_info *info)
 			goto out;
 	}
 
-	if (info->attrs[WGDEVICE_A_FLAGS] &&
-	    nla_get_u32(info->attrs[WGDEVICE_A_FLAGS]) &
-		    WGDEVICE_F_REPLACE_PEERS)
+	if (flags & WGDEVICE_F_REPLACE_PEERS)
 		wg_peer_remove_all(wg);
 
 	if (info->attrs[WGDEVICE_A_PRIVATE_KEY] &&
@@ -532,6 +553,10 @@ static int wg_set_device(struct sk_buff *skb, struct genl_info *info)
 		u8 public_key[NOISE_PUBLIC_KEY_LEN];
 		struct wg_peer *peer, *temp;
 
+		if (!crypto_memneq(wg->static_identity.static_private,
+				   private_key, NOISE_PUBLIC_KEY_LEN))
+			goto skip_set_private_key;
+
 		/* We remove before setting, to prevent race, which means doing
 		 * two 25519-genpub ops.
 		 */
@@ -549,12 +574,15 @@ static int wg_set_device(struct sk_buff *skb, struct genl_info *info)
 							 private_key);
 		list_for_each_entry_safe(peer, temp, &wg->peer_list,
 					 peer_list) {
-			if (!wg_noise_precompute_static_static(peer))
+			if (wg_noise_precompute_static_static(peer))
+				wg_noise_expire_current_peer_keypairs(peer);
+			else
 				wg_peer_remove(peer);
 		}
 		wg_cookie_checker_precompute_device_keys(&wg->cookie_checker);
 		up_write(&wg->static_identity.lock);
 	}
+skip_set_private_key:
 
 	if (info->attrs[WGDEVICE_A_PEERS]) {
 		struct nlattr *attr, *peer[WGPEER_A_MAX + 1];
